
/*****************************************************************************
 *                                                                           *
 * This FILE psu120_lib.c                                                    *
 * is part of the PSU120 library.  It is based on code copyright 1996 and    *
 * 1998 by Adimec Advanced Image Systems B.V.  Unix conversion copyright     *
 * 1999 and 2000 by Gregory L. Kreider.  All rights reserved.  See below     *
 * for distribution terms.                                                   *
 *                                                                           *
 * The code in this file provides a high low interface for programming       *
 * a PSU120/MX12P camera.  It follows the specification in the Adimec        *
 * help files.  The PSU120 library communicates with the serial port and     *
 * composes/decodes messages to/from the PSU120.  The MX12P library should   *
 * be used to program the camera, as it is a higher-level set of functions.  *
 *                                                                           *
 * See libraries.doc for porting notes.                                      *
 *                                                                           *
 * This software package is provided 'as is' without any further warranties  *
 * express or implied beyond those provided for by local law.  Neither       *
 * the authors nor any distributors are liable for damages, special,         *
 * indirect, or consequential, incurred by the use or inability to use this  *
 * software, or for any claim by any other party.                            *
 *                                                                           *
 * No further restrictions are placed on the use or distribution of this     *
 * code.                                                                     *
 *                                                                           *
 *****************************************************************************/




#include <stdio.h>

/* need to define this to get usleep from unistd.h */
/* we can do this because we don't use any of the other BSD-type functions */
#define __USE_BSD
#include <unistd.h>
#undef __USE_BSD

#include <fcntl.h>
#include <termios.h>
#include <errno.h>
#include "psu120_lib.h"


#define PSU_BUF_SIZE       33    /* maximum buffer size */
#define PSU_DAT_SIZE     33-4    /* data size: msg - cmd - p1 - p2 - crc */

/* PSU120 communication control characters */
#define ccACK               6    /* positive acknowledge */
#define ccNAK              21    /* negative acknowledge */
#define ccSTX               2    /* message start */
#define ccETX               3    /* message end */
#define ccSUB              26    /* substitute next byte */

/* PSU120 command codes */
#define ccATT            0xFF    /* Attention */
#define ccVER            0xFE    /* Version */
#define ccWR             0xFD    /* Write */
#define ccRD             0xFC    /* Read */
#define ccWRNV           0xFB    /* Write to NVRAM */
#define ccRDNV           0xFA    /* Read from NVRAM */

/* PSU120 NVRAM slave address */
#define NVRAM_I2C_ADDR   0x50

/* comm parameters */
#define TxTO                2    /* 2 second transmitter timeout */
#define RxTO                2    /* 2 second receiver timeout */
#define RxRC                6    /* number retries (*2) after failure */

/* local error codes */
/* these conflict with ERR_PSU, so shift to 2x */
#define ERR_INT_RXTO      -21    /* receiver timeout */
#define ERR_INT_ICRC      -22    /* crc check failed */
#define ERR_INT_SIZE      -23    /* receiver buffer overflow */
#define ERR_INT_COMM      -24    /* comm error (b.v. no ACK) */

/* local structures */

/* Generic PSU120 message (send or receive) */
struct TCmd_t {
  uchar ccCmd;          /* PSU120 command code */
  int UseParam;         /* use extra parameter */
  uchar Param;          /* value of extra parameter */
  uchar TxDataSize;     /* length of data section */
  uchar RxDataSize;     /* length of reply message */
  uchar* TxData;        /* data */
  uchar* RxData;        /* reply */
};

typedef struct TCmd_t TCmd;

struct termios old_portinfo;


/* local functions */
int RemoteExecuteCmd(int, TCmd*);
void TransmitCmd(int, TCmd*);
int ReceivePSUMessage(int, uchar*, uchar*);
int ReceiveByte(int, uchar*);
void HandleSubstitution(uchar*, uchar, int*);

/* no equivalent of DllEntryPoint */




/***************************************************************
 ***************************************************************
 * 
 *  Public Functions
 *
 ***************************************************************
 ***************************************************************/


/***************************************************************
 * 
 * PSUOPEN:
 *   Open the serial port /dev/ttySx (where x is one less than
 *   the serial port number, dwz 0 for Comm Port 1) and set up
 *   the terminal io structure.
 * Returns:
 *   File descriptor of open connection if successfully opened,
 *   or an error
 *
 ***************************************************************/

int PSUOPEN(int commport)
{
  int i;
  int fd;
  struct termios portinfo;
  char devname[10];

  /* commport is already the device number */
  if ((commport < PSU_COM1) || (commport > PSU_COM8))
	return(ERR_PSU_PORT);

  /* note: assumes 1 digit comm port number */
  sprintf(devname,"/dev/ttyS%d",commport);
  if ((fd = open(devname,O_RDWR | O_NOCTTY)) < 0)
	return(ERR_PSU_OPEN);

  if (tcflush(fd,TCIOFLUSH) < 0)
	return(ERR_PSU_OPEN);

  /* from windows code: 8 bits, no parity, stop bit, 4800 baud
	                    no RTS, no DTR, no XON/XOFF */
  /* no timeout; use microsleep later on */
  portinfo.c_iflag = 0;
  portinfo.c_oflag = 0;
  portinfo.c_cflag = CS8 | CREAD | CLOCAL | B4800;
  portinfo.c_lflag = 0;
  for (i=0; i<NCCS; i++)
	portinfo.c_cc[i] = 0;

  if (tcgetattr(fd,&old_portinfo) < 0)
	return(ERR_PSU_OPEN);
  if (tcsetattr(fd,TCSANOW,&portinfo) < 0)
	return(ERR_PSU_OPEN);

  return(fd);
}


/***************************************************************
 * 
 * PSUATTENTION:
 *   Send an ATT command to the PSU120 and wait for 
 *   acknowledgement.
 * Returns:
 *   Result of command (0 on succeed, or error code)
 *
 ***************************************************************/

int PSUATTENTION(int fd)
{
  TCmd psucmd;

  psucmd.ccCmd = ccATT;
  psucmd.UseParam = psucmd.Param = 0;
  psucmd.TxDataSize = psucmd.RxDataSize = 0;
  psucmd.TxData = psucmd.RxData = NULL;

  return(RemoteExecuteCmd(fd,&psucmd));
}


/***************************************************************
 * 
 * PSUVERSION:
 *   Get the version string from the PSU and store it in the
 *   character array (which much be large enough to hold the
 *   result; use PSU_DATA_SIZE+1).
 * Returns:
 *   Result of command (0 on succeed, or error code).  String
 *   is stored in array.
 *
 ***************************************************************/

int PSUVERSION(int fd, char* verstr)
{
  TCmd psucmd;
  int err;

  psucmd.ccCmd = ccVER;
  psucmd.UseParam = psucmd.Param = 0;
  psucmd.TxDataSize = psucmd.RxDataSize = 0;
  psucmd.TxData = NULL;
  psucmd.RxData = (char *) verstr;

  err = RemoteExecuteCmd(fd,&psucmd);
  if (err == 0) verstr[psucmd.RxDataSize] = 0;
  return(err);
}


/***************************************************************
 * 
 * PSUWRITE:
 *   Write data (len characters stored in the data array) to
 *   the I2C bus (at address addr).
 * Returns:
 *   Result of command (0 on succeed, or error code).
 *
 ***************************************************************/

int PSUWRITE(int fd, uchar addr, uchar len, uchar* data)
{
  TCmd psucmd;

  if (len > PSU_DAT_SIZE)
	return(ERR_PSU_SIZE);

  psucmd.ccCmd = ccWR;
  psucmd.UseParam = 1;
  psucmd.Param = addr;
  psucmd.TxDataSize = len;
  psucmd.TxData = (uchar *) data;
  psucmd.RxDataSize = 0;
  psucmd.RxData = NULL;

  return(RemoteExecuteCmd(fd,&psucmd));
}


/***************************************************************
 * 
 * PSUREAD:
 *   Get data (len bytes from I2C address addr, stored in data
 *   array) from PSU.
 * Returns:
 *   Result of command (0 on succeed, or error code).
 *
 ***************************************************************/

int PSUREAD(int fd, uchar addr, uchar len, uchar* data)
{
  TCmd psucmd;

  if (len > PSU_DAT_SIZE)
	return(ERR_PSU_SIZE);

  psucmd.ccCmd = ccRD;
  psucmd.UseParam = 1;
  psucmd.Param = addr;
  psucmd.TxDataSize = 0;
  psucmd.TxData = NULL;
  psucmd.RxDataSize = len;
  psucmd.RxData = data;

  return(RemoteExecuteCmd(fd,&psucmd));
}


/***************************************************************
 * 
 * PSUWRITENV:
 *   Send data array to NVRAM.  Send byte by byte to delay 
 *   writes.
 * Returns:
 *   Result of command (0 on succeed, or error code).
 *
 ***************************************************************/

int PSUWRITENV(int fd, uchar addr, uchar len, uchar* data)
{
  int i;
  TCmd psucmd;
  int err;
  
  if (len > PSU_DAT_SIZE)
	return(ERR_PSU_SIZE);

  err = 0;
  for (i=0; i<len; i++) {
	psucmd.ccCmd = ccWRNV;
	psucmd.UseParam = 1;
	psucmd.Param = (uchar) (addr + (uchar) i);
	psucmd.TxDataSize = 1;
	psucmd.TxData = &data[i];
	psucmd.RxDataSize = 0;
	psucmd.RxData = NULL;

	err = RemoteExecuteCmd(fd,&psucmd);
	if (err < 0)
	  return(err);
  }

  return(err);
}


/***************************************************************
 * 
 * PSUREADENV:
 *   Get data array from NVRAM.  Read as block.  If fails at
 *   given address, try NVRAM address.
 * Returns:
 *   Result of command (0 on succeed, or error code).
 *
 ***************************************************************/

int PSUREADENV(int fd, uchar addr, uchar len, uchar* data)
{
  uchar inchar;
  TCmd psucmd;
  int err;

  if (len > PSU_DAT_SIZE)
	return(ERR_PSU_SIZE);

  psucmd.ccCmd = ccRDNV;
  psucmd.UseParam = 1;
  psucmd.Param = addr;
  psucmd.TxDataSize = 0;
  psucmd.TxData = NULL;
  psucmd.RxDataSize = len;
  psucmd.RxData = data;
  err = RemoteExecuteCmd(fd,&psucmd);
  if (err < 0) {
	inchar = addr;
	err = PSUWRITE(fd,NVRAM_I2C_ADDR,1,&inchar);
	if (err < 0)
	  return(err);
	err = PSUREAD(fd,NVRAM_I2C_ADDR,len,data);
  }
  return(err);
}


/***************************************************************
 * 
 * PSUCLOSE:
 *   Terminate connection to serial device and clear terminal
 *   data.
 * Returns:
 *   Result of command (0 on succeed, or error code).
 *
 ***************************************************************/

int PSUCLOSE(int fd)
{
  int err;

  if ((err = tcsetattr(fd,TCSANOW,&old_portinfo)) < 0)
	return(err);
  return(close(fd));
}


/***************************************************************
 * 
 * PSUSTARTPOLL:
 *   Send an ATTENTION signal without waiting for a reply.
 * Returns:
 *   0.
 *
 ***************************************************************/

int PSUSTARTPOLL(int fd)
{
  TCmd psucmd;

  psucmd.ccCmd = ccATT;
  psucmd.UseParam = psucmd.Param = 0;
  psucmd.TxDataSize = psucmd.RxDataSize = 0;
  psucmd.TxData = psucmd.RxData = NULL;
  TransmitCmd(fd,&psucmd);
  return(0);
}


/***************************************************************
 * 
 * PSUSTARTPOLL:
 *   Check for an acknowledgement (preserve any existing termio
 *   data while reading the port).
 * Returns:
 *   1 on success, or error code.
 *
 ***************************************************************/

int PSUPOLL(int fd)
{
  uchar inchar;
  struct termios portinfo,originfo;
  
  tcflush(fd,TCIOFLUSH);
  if (tcgetattr(fd,&originfo) < 0)
	return(ERR_PSU_COMM);

  tcgetattr(fd,&portinfo);
  portinfo.c_cc[VMIN] = 0;
  portinfo.c_cc[VTIME] = 1;
  tcflush(fd,TCIOFLUSH);
  if (tcsetattr(fd,TCSANOW,&portinfo) < 0)
	return(ERR_PSU_COMM);

  if (read(fd,&inchar,1) != 1)
	return(ERR_PSU_COMM);

  tcsetattr(fd,TCSANOW,&originfo);

  if (inchar == (uchar) ccACK)
	return(1);
  return(ERR_PSU_COMM);
}
  


/***************************************************************
 ***************************************************************
 * 
 *  Private Functions
 *
 ***************************************************************
 ***************************************************************/


/***************************************************************
 * 
 * RemoteExecuteCommand:
 *   Repeatedly try to send the command to the PSU and get a
 *   reply.  The macro RxRC defines the Number of retries.
 * Returns:
 *   0 on success, or < 0 as an error code.
 *
 ***************************************************************/

int RemoteExecuteCmd(int fd, TCmd* psucmd)
{
  int retry;
  uchar reply;
  int err;

  retry = RxRC + 1;
  err = 0;
  do {
	TransmitCmd(fd,psucmd);
	if ((tcdrain(fd)) < 0) {
	  err = ERR_INT_COMM;
	  break;
	}
	reply = 0;
	ReceiveByte(fd,&reply);
	if (reply == ccACK) {
	  if (psucmd->RxData != NULL)
		err = ReceivePSUMessage(fd,&(psucmd->RxDataSize),psucmd->RxData);
	  else err = 0;
	}
	else
	  err = ERR_INT_COMM;
	if (err < 0) retry--;
  } while ((err < 0) && (retry > 0));
  return(err);
}


/***************************************************************
 * 
 * TransmitCmd:
 *   Send a command to the PSU by setting up an array of bytes
 *   to send, tracking the checksum and substituting two 
 *   characters for one where needed.  Then, write it in one go.
 * Returns:
 *   nothing.
 *
 ***************************************************************/

void TransmitCmd(int fd, TCmd* psucmd)
{
  int i,index;
  uchar TxBuffer[PSU_BUF_SIZE*2+5];
  uchar crc;

  index = 0;
  TxBuffer[index++] = ccSTX;
  crc = (uchar) 0 ^ psucmd->ccCmd;
  HandleSubstitution(TxBuffer,psucmd->ccCmd,&index);
  if (psucmd->UseParam) {
	crc ^= psucmd->Param;
	HandleSubstitution(TxBuffer,psucmd->Param,&index);
  }
  if ((psucmd->RxData != NULL) && (psucmd->RxDataSize != 0)) {
	crc ^= psucmd->RxDataSize;
	HandleSubstitution(TxBuffer,psucmd->RxDataSize,&index);
  }
  if (psucmd->TxData != NULL) {
	for (i=0; i<psucmd->TxDataSize; i++) {
	  crc ^= psucmd->TxData[i];
	  HandleSubstitution(TxBuffer,psucmd->TxData[i],&index);
	}
  }
  HandleSubstitution(TxBuffer,crc,&index);
  TxBuffer[index++] = ccETX;

  /* should check write here */
  write(fd,TxBuffer,index);

  usleep(10000L);
}


/***************************************************************
 * 
 * HandleSubstitution:
 *   Encode byte values less than 32 following the I2C spec.
 * Returns:
 *   nothing
 *
 ***************************************************************/

void HandleSubstitution(uchar* TxBuffer, uchar data, int* index)
{
  if (data < 32) {
	TxBuffer[*index] = ccSUB;
	(*index)++;
	TxBuffer[*index] = (uchar) (data ^ (uchar) 0xff);
	(*index)++;
  }
  else {
	TxBuffer[*index] = data;
	(*index)++;
  }
}


/***************************************************************
 * 
 * ReceivePSUMessage:
 *   Wait for an STX code to start a transmission from the PSU,
 *   then read until ETX, tracking the checksum and unpacking
 *   any substitutions.  Return errors on a bad checksum, a
 *   message that is too large.  (Note that timeout errors won't
 *   happen under Unix, even though we test for the 
 *   possibility).
 * Returns:
 *   0 on success, or < 0 on error.
 *
 ***************************************************************/

int ReceivePSUMessage(int fd, uchar* len, uchar* data)
{
  int index;
  int SUBNext;
  uchar rxd,crc;
  int err;

  /* get STX to mark start of message */
  do {
	err = ReceiveByte(fd,&rxd);
  } while ((rxd != ccSTX) && (err >= 0));
  /* this will never happen */
  if (err < 0)
	return(ERR_INT_RXTO);
  
  /* now read message */
  SUBNext = index = crc = 0;
  do {
	err = ReceiveByte(fd,&rxd);
	/* PSU_DAT_SIZE+2 for STX and ETX */
	if ((err >= 0) && (rxd != ccETX) && (index < (PSU_DAT_SIZE+2))) {
	  if (rxd == ccSUB) 
		SUBNext = 1;
	  else {
		if (SUBNext) {
		  data[index] = (uchar) (rxd ^ (uchar) 0xff);
		  SUBNext = 0;
		}
		else data[index] = rxd;
		crc ^= data[index];
		index++;
	  }
	}
  } while ((rxd != ccETX) && (err >= 0));
  
  /* timeout? this will never happen */
  if (err < 0)
	return(ERR_INT_RXTO);
  /* corrupt transmission? */
  if (crc != 0)
	return(ERR_INT_ICRC);
  /* buffer overflow? */
  if (index == (PSU_DAT_SIZE+2))
	return(ERR_INT_SIZE);
  /* well, no! */
  *len = (uchar) (index-1);
  return(0);
}


/***************************************************************
 * 
 * ReceiveByte:
 *   Read one byte from the serial file and sleep 10 ms before
 *   getting the next.
 * Returns:
 *   0 on success, or < 0 on error.
 *
 ***************************************************************/

int ReceiveByte(int fd, uchar* data)
{
  int status;

  if ((status = read(fd,data,1)) < 0) {
	return(status);
  }
  /* wait 10 ms before reading --- see timing info in Adimec docs */
  usleep(10000);
  return(0);
}

